==> lith-1.3.1.tar.gz: Lith-1.3.1/src/weechat.cpp <==
// Lith
// Copyright (C) 2020 Martin Bříza
// Copyright (C) 2020 Václav Kubernát
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; If not, see <http://www.gnu.org/licenses/>.

#include "weechat.h"

#include "lith.h"
#include "protocol.h"

#include <QThread>

#include <QPasswordDigestor>
#include <QCryptographicHash>
#include <QRandomGenerator>

Weechat::Weechat(Lith *lith)
    : QObject(nullptr)
    , m_connection(new SocketHelper(this))
    , m_lith(lith)
{
    connect(m_connection, &SocketHelper::dataReceived, this, &Weechat::onDataReceived, Qt::QueuedConnection);
    connect(m_connection, &SocketHelper::connected, this, &Weechat::onConnected, Qt::QueuedConnection);
    connect(m_connection, &SocketHelper::disconnected, this, &Weechat::onDisconnected, Qt::QueuedConnection);
    connect(m_connection, &SocketHelper::errorOccurred, this, &Weechat::onError, Qt::QueuedConnection);

    connect(lith, &Lith::pongReceived, this, &Weechat::onPongReceived, Qt::QueuedConnection);
    connect(m_pingTimer, &QTimer::timeout, this, &Weechat::onPingTimeout, Qt::QueuedConnection);
    m_pingTimer->setSingleShot(false);
    m_pingTimer->start(5000);

    connect(m_reconnectTimer, &QTimer::timeout, this, &Weechat::restart, Qt::QueuedConnection);
    m_reconnectTimer->setInterval(100);
    m_reconnectTimer->setSingleShot(false);
}

Lith *Weechat::lith() {
    return m_lith;
}

const QStringList supportedHashAlgos {
    "plain",
    "sha256",
    "sha512",
    "pbkdf2+sha256",
    "pbkdf2+sha512"
};
QByteArray Weechat::hashPassword(const QString &password, const QString &algo, const QByteArray &salt, int iterations) {
    if (algo == "plain") {
        return password.toUtf8();
    }
    else if (algo == "sha256") {
        QCryptographicHash qch(QCryptographicHash::Sha256);
        qch.addData(salt);
        qch.addData(password.toUtf8());
        return qch.result();
    }
    else if (algo == "sha512") {
        QCryptographicHash qch(QCryptographicHash::Sha512);
        qch.addData(salt);
        qch.addData(password.toUtf8());
        return qch.result();
    }
    else if (algo == "pbkdf2+sha256") {
        return QPasswordDigestor::deriveKeyPbkdf2(QCryptographicHash::Sha256, password.toUtf8(), salt, iterations, 32);
    }
    else if (algo == "pbkdf2+sha512") {
        return QPasswordDigestor::deriveKeyPbkdf2(QCryptographicHash::Sha512, password.toUtf8(), salt, iterations, 64);
    }
    return {};
}

QByteArray Weechat::randomString(int length) {
    QByteArray alphabet { "0123456789ABCDEF"};
    QByteArray result;
    auto rg = QRandomGenerator::securelySeeded();
    for (int i = 0; i < length; i++) {
        result += alphabet[rg.bounded(0, alphabet.count())];
    }
    return result;
}

void Weechat::init() {
    m_timeoutTimer->setInterval(5000);
    m_timeoutTimer->setSingleShot(true);
    //connect(m_timeoutTimer, &QTimer::timeout, this, &Weechat::onTimeout, Qt::QueuedConnection);

    connect(m_hotlistTimer, &QTimer::timeout, this, &Weechat::requestHotlist, Qt::QueuedConnection);
    m_hotlistTimer->setInterval(10000);
    m_hotlistTimer->setSingleShot(false);

    connect(lith()->settingsGet(), &Settings::ready, this, &Weechat::onConnectionSettingsChanged, Qt::QueuedConnection);
    connect(lith()->settingsGet(), &Settings::hostChanged, this, &Weechat::onConnectionSettingsChanged, Qt::QueuedConnection);
    connect(lith()->settingsGet(), &Settings::passphraseChanged, this, &Weechat::onConnectionSettingsChanged, Qt::QueuedConnection);
    connect(lith()->settingsGet(), &Settings::portChanged, this, &Weechat::onConnectionSettingsChanged, Qt::QueuedConnection);
    connect(lith()->settingsGet(), &Settings::encryptedChanged, this, &Weechat::onConnectionSettingsChanged, Qt::QueuedConnection);

    onConnectionSettingsChanged();
}

void Weechat::start() {
    m_connection->reset();
    m_restarting = false;
    qCritical() << "Connecting";

    lith()->statusSet(Lith::CONNECTING);

    restart();
}

void Weechat::restart() {
    m_initializationStatus = UNINITIALIZED;
    auto host = lith()->settingsGet()->hostGet();
    auto port = lith()->settingsGet()->portGet();
    auto ssl = lith()->settingsGet()->encryptedGet();
    auto websocketEndpoint = lith()->settingsGet()->websocketsEndpointGet();
#ifndef __EMSCRIPTEN__
    if (!lith()->settingsGet()->useWebsocketsGet())
        m_connection->connectToTcpSocket(host, port, ssl);
    else // BEWARE
#endif // __EMSCRIPTEN__
        m_connection->connectToWebsocket(host, websocketEndpoint, port, ssl);
    // BEWARE OF THE ELSE ABOVE
}

void Weechat::onConnectionSettingsChanged() {
    auto host = lith()->settingsGet()->hostGet();
    auto pass = lith()->settingsGet()->passphraseGet();
    if (!host.isEmpty() && !pass.isEmpty()) {
        qCritical() << "CONNECTING";
        m_connection->reset();
        if (!m_restarting)
            QTimer::singleShot(50, this, &Weechat::start);
        m_restarting = true;
    }
}

void Weechat::onHandshakeAccepted(const StringMap &data) {
    if (!m_connection->isConnected())
        return;

    auto algo = data["password_hash_algo"];
    auto iterations = data["password_hash_iterations"].toInt();
    auto serverNonce = QByteArray::fromHex(data["nonce"].toLocal8Bit());
    auto clientNonce = QByteArray::fromHex(randomString(16));
    auto pass = lith()->settingsGet()->passphraseGet();

    auto salt = serverNonce + clientNonce;
    auto hash = hashPassword(pass, algo, salt, iterations);

    QString hashString;
    if (algo == "plain")
        hashString = "password=" + pass + ",compression=" + (lith()->settingsGet()->connectionCompressionGet() ? "zlib" : "off");
    else if (algo.startsWith("pbkdf2"))
        hashString = "password_hash=" + algo + ':' + salt.toHex() + ':' + QString("%1").arg(iterations) + ':' + hash.toHex();
    else
        hashString = "password_hash=" + algo + ':' + salt.toHex() + ':' + hash.toHex();

    m_initializationStatus = (Initialization) (m_initializationStatus | HANDSHAKE);

    m_connection->write(("init " + hashString + "\n").toUtf8());
    m_connection->write(QString("(%1) hdata buffer:gui_buffers(*) number,name,short_name,hidden,title,local_variables\n").arg(MessageNames::c_requestBuffers).toUtf8());
    m_connection->write(QString("(%1) hdata buffer:gui_buffers(*)/lines/last_line(-1)/data\n").arg(MessageNames::c_requestFirstLine).toUtf8());
    m_connection->write(QString("(%1) hdata hotlist:gui_hotlist(*)\n").arg(MessageNames::c_requestHotlist).toUtf8());
    m_connection->write("sync\n");
    m_connection->write(QString("(%1) nicklist\n").arg(MessageNames::c_requestNicklist).toUtf8());
}

void Weechat::requestHotlist() {
    if (m_connection->isConnected()) {
        auto msg = QString("(handleHotlist;%1) hdata hotlist:gui_hotlist(*)\n").arg(m_messageOrder++);
        m_connection->write(msg.toUtf8());
    }
}


void Weechat::onConnected() {
    qCritical() << "Connected!";

    m_reconnectTimer->stop();
    m_reconnectTimer->setInterval(100);

    QTimer::singleShot(0, lith(), &Lith::resetData);
    lith()->networkErrorStringSet(QString());

    lith()->statusSet(Lith::CONNECTED);
    QString hashAlgos;
    for (auto &i : supportedHashAlgos) {
        if (!hashAlgos.isEmpty())
            hashAlgos.append(":");
        hashAlgos.append(i);
    }

    if (lith()->settingsGet()->handshakeAuthGet()) {
        m_connection->write(QString("(%1) handshake password_hash_algo=%2,compression=%3\n").arg(MessageNames::c_handshake).arg(hashAlgos).arg(lith()->settingsGet()->connectionCompressionGet() ? "zlib" : "off").toUtf8());
    }
    else {
        StringMap data;
        data["password_hash_algo"] = "plain";
        onHandshakeAccepted(data);
    }

    m_hotlistTimer->start();
}

void Weechat::onDisconnected() {
    lith()->statusSet(Lith::DISCONNECTED);

    m_fetchBuffer.clear();
    m_bytesRemaining = 0;
    m_hotlistTimer->stop();

    m_reconnectTimer->setInterval(m_reconnectTimer->interval() * 2);
    m_reconnectTimer->start();
}

void Weechat::onDataReceived(const QByteArray &data) {
    auto dataCopy(data);
    onMessageReceived(dataCopy);
}

void Weechat::onError(const QString &message) {
    lith()->statusSet(Lith::ERROR);
    lith()->networkErrorStringSet("Connection failed: "+ message);
}

bool Weechat::input(pointer_t ptr, const QString &data) {
    // server doesn't reply to input commands directly so no message order here
    auto line = QString("input 0x%1 %2\n").arg(ptr, 0, 16).arg(data);
    //qCritical() << "WRITING:" << line;
    auto message = line.toUtf8();
    if (message.count() == m_connection->write(line.toUtf8()))
        return true;
    return false;
}

void Weechat::fetchLines(pointer_t ptr, int count) {
    auto line = QString("(handleFetchLines;%1) hdata buffer:0x%2/lines/last_line(-%3)/data\n").arg(m_messageOrder++).arg(ptr, 0, 16).arg(count);
    //qCritical() << "WRITING:" << line;
    m_connection->write(line.toUtf8());
    m_timeoutTimer->start(5000);
}

void Weechat::onMessageReceived(QByteArray &data) {
    //qCritical() << "Message!" << data;
    QDataStream s(&data, QIODevice::ReadOnly);

    Protocol::String id = Protocol::parse<Protocol::String>(s);

    char type[4] = { 0 };
    s.readRawData(type, 3);

    if (QString(type) == "hda") {
        Protocol::HData hda = Protocol::parse<Protocol::HData>(s);

        if (c_initializationMap.contains(id)) {
            // wtf, why can't I write this as |= ?
            m_initializationStatus = (Initialization) (m_initializationStatus | c_initializationMap.value(id, UNINITIALIZED));
            if (!QMetaObject::invokeMethod(Lith::instance(), id.toStdString().c_str(), Qt::QueuedConnection, Q_ARG(Protocol::HData, hda))) {
                qWarning() << "Possible unhandled message:" << id;
            }
        }
        else {
            auto name = id.split(";").first();
            if (!QMetaObject::invokeMethod(Lith::instance(), name.toStdString().c_str(), Qt::QueuedConnection, Q_ARG(Protocol::HData, hda))) {
                qWarning() << "Possible unhandled message:" << name;
            }
        }
    }
    else if (QString(type) == "htb") {
        Protocol::HashTable htb = Protocol::parse<Protocol::HashTable>(s);

        onHandshakeAccepted(htb);
    }
    else if (QString(type) == "str") {
        Protocol::String str = Protocol::parse<Protocol::String>(s);

        if (!QMetaObject::invokeMethod(Lith::instance(), id.toStdString().c_str(), Qt::QueuedConnection, Q_ARG(const FormattedString&, str))) {
            qWarning() << "Possible unhandled message:" << id;
        }
    }
    else {
        qCritical() << "onMessageReceived is not handling type: " << type;
    }

    if (!s.atEnd()) {
        qCritical() << "STREAM WAS NOT AT END!!!";
    }
}

void Weechat::onPongReceived(qint64 id) {
    m_lastReceivedPong = id;
}

void Weechat::onTimeout() {
    m_connection->reset();
    lith()->statusSet(Lith::DISCONNECTED);
    start();
}

void Weechat::onPingTimeout() {
    static qint64 previousPing = 0;
    if (m_initializationStatus == COMPLETE) {
        if (previousPing < m_lastReceivedPong - 1) {
            restart();
        }
        previousPing = m_messageOrder++;
        if (m_connection->write(QString("(%1) ping %1\n").arg(previousPing)) <= 0) {
            restart();
        }
    }
    else {
        //restart();
    }
}

==> lith-1.3.34.tar.gz: Lith-1.3.34/src/weechat.cpp <==
// Lith
// Copyright (C) 2020 Martin Bříza
// Copyright (C) 2020 Václav Kubernát
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; If not, see <http://www.gnu.org/licenses/>.

#include "weechat.h"

#include "lith.h"
#include "protocol.h"

#include <QThread>

#include <QPasswordDigestor>
#include <QCryptographicHash>
#include <QRandomGenerator>

Weechat::Weechat(Lith *lith)
    : QObject(nullptr)
    , m_connection(new SocketHelper(this))
    , m_lith(lith)
{
    connect(m_connection, &SocketHelper::dataReceived, this, &Weechat::onDataReceived, Qt::QueuedConnection);
    connect(m_connection, &SocketHelper::connected, this, &Weechat::onConnected, Qt::QueuedConnection);
    connect(m_connection, &SocketHelper::disconnected, this, &Weechat::onDisconnected, Qt::QueuedConnection);
    connect(m_connection, &SocketHelper::errorOccurred, this, &Weechat::onError, Qt::QueuedConnection);

    connect(lith, &Lith::pongReceived, this, &Weechat::onPongReceived, Qt::QueuedConnection);
    connect(m_pingTimer, &QTimer::timeout, this, &Weechat::onPingTimeout, Qt::QueuedConnection);
    m_pingTimer->setSingleShot(false);
    m_pingTimer->start(5000);

    connect(m_reconnectTimer, &QTimer::timeout, this, &Weechat::restart, Qt::QueuedConnection);
    m_reconnectTimer->setInterval(100);
    m_reconnectTimer->setSingleShot(false);
}

const Lith *Weechat::lith() const {
    return m_lith;
}

Lith *Weechat::lith() {
    return m_lith;
}

const QStringList supportedHashAlgos {
    "plain",
    "sha256",
    "sha512",
    "pbkdf2+sha256",
    "pbkdf2+sha512"
};
QByteArray Weechat::hashPassword(const QString &password, const QString &algo, const QByteArray &salt, int iterations) {
    if (algo == "plain") {
        return password.toUtf8();
    }
    else if (algo == "sha256") {
        QCryptographicHash qch(QCryptographicHash::Sha256);
        qch.addData(salt);
        qch.addData(password.toUtf8());
        return qch.result();
    }
    else if (algo == "sha512") {
        QCryptographicHash qch(QCryptographicHash::Sha512);
        qch.addData(salt);
        qch.addData(password.toUtf8());
        return qch.result();
    }
    else if (algo == "pbkdf2+sha256") {
        return QPasswordDigestor::deriveKeyPbkdf2(QCryptographicHash::Sha256, password.toUtf8(), salt, iterations, 32);
    }
    else if (algo == "pbkdf2+sha512") {
        return QPasswordDigestor::deriveKeyPbkdf2(QCryptographicHash::Sha512, password.toUtf8(), salt, iterations, 64);
    }
    return {};
}

QByteArray Weechat::randomString(int length) {
    QByteArray alphabet { "0123456789ABCDEF"};
    QByteArray result;
    auto rg = QRandomGenerator::securelySeeded();
    for (int i = 0; i < length; i++) {
        result += alphabet[rg.bounded(0, alphabet.count())];
    }
    return result;
}

void Weechat::init() {
    m_timeoutTimer->setInterval(5000);
    m_timeoutTimer->setSingleShot(true);
    //connect(m_timeoutTimer, &QTimer::timeout, this, &Weechat::onTimeout, Qt::QueuedConnection);

    connect(m_hotlistTimer, &QTimer::timeout, this, &Weechat::requestHotlist, Qt::QueuedConnection);
    m_hotlistTimer->setInterval(10000);
    m_hotlistTimer->setSingleShot(false);

    connect(lith()->settingsGet(), &Settings::ready, this, &Weechat::onConnectionSettingsChanged, Qt::QueuedConnection);
    connect(lith()->settingsGet(), &Settings::networkSettingsChanged, this, &Weechat::onConnectionSettingsChanged, Qt::QueuedConnection);

    onConnectionSettingsChanged();
}

void Weechat::start() {
    m_connection->reset();
    m_restarting = false;
    lith()->statusSet(Lith::CONNECTING);
    restart();
}

void Weechat::restart() {
    m_initializationStatus = UNINITIALIZED;
    auto host = lith()->settingsGet()->hostGet();
    auto port = lith()->settingsGet()->portGet();
    auto ssl = lith()->settingsGet()->encryptedGet();
    auto websocketEndpoint = lith()->settingsGet()->websocketsEndpointGet();
#ifndef __EMSCRIPTEN__
    if (!lith()->settingsGet()->useWebsocketsGet())
        m_connection->connectToTcpSocket(host, port, ssl);
    else // BEWARE
#endif // __EMSCRIPTEN__
        m_connection->connectToWebsocket(host, websocketEndpoint, port, ssl);
    // BEWARE OF THE ELSE ABOVE
}

void Weechat::onConnectionSettingsChanged() {
    auto host = lith()->settingsGet()->hostGet();
    auto pass = lith()->settingsGet()->passphraseGet();
    lith()->log(Logger::Network, "Connection settings have changed, will reset the socket");
    if (!host.isEmpty() && !pass.isEmpty()) {
        m_connection->reset();
        if (!m_restarting)
            QTimer::singleShot(50, this, &Weechat::start);
        m_restarting = true;
    }
}

void Weechat::onHandshakeAccepted(const StringMap &data) {
    if (!m_connection->isConnected())
        return;

    lith()->log(Logger::Protocol, "WeeChat handshake was accepted, authenticating");

    auto algo = data["password_hash_algo"];
    auto iterations = data["password_hash_iterations"].toInt();
    auto serverNonce = QByteArray::fromHex(data["nonce"].toLocal8Bit());
    auto clientNonce = QByteArray::fromHex(randomString(16));
    auto pass = lith()->settingsGet()->passphraseGet();

    auto salt = serverNonce + clientNonce;
    auto hash = hashPassword(pass, algo, salt, iterations);

    QString hashString;
    if (algo == "plain")
        hashString = "password=" + pass + ",compression=" + (lith()->settingsGet()->connectionCompressionGet() ? "zlib" : "off");
    else if (algo.startsWith("pbkdf2"))
        hashString = "password_hash=" + algo + ':' + salt.toHex() + ':' + QString("%1").arg(iterations) + ':' + hash.toHex();
    else
        hashString = "password_hash=" + algo + ':' + salt.toHex() + ':' + hash.toHex();

    m_initializationStatus = (Initialization) (m_initializationStatus | HANDSHAKE);

    m_connection->write(("init " + hashString + "\n").toUtf8());
    m_connection->write(QString("(%1) hdata buffer:gui_buffers(*) number,name,short_name,hidden,title,local_variables\n").arg(MessageNames::c_requestBuffers).toUtf8());
    m_connection->write(QString("(%1) hdata buffer:gui_buffers(*)/lines/last_line(-1)/data\n").arg(MessageNames::c_requestFirstLine).toUtf8());
    m_connection->write(QString("(%1) hdata hotlist:gui_hotlist(*)\n").arg(MessageNames::c_requestHotlist).toUtf8());
    m_connection->write("sync\n");
    m_connection->write(QString("(%1) nicklist\n").arg(MessageNames::c_requestNicklist).toUtf8());
}

void Weechat::requestHotlist() {
    if (m_connection->isConnected()) {
        auto msg = QString("(handleHotlist;%1) hdata hotlist:gui_hotlist(*)\n").arg(m_messageOrder++);
        m_connection->write(msg.toUtf8());
    }
}


void Weechat::onConnected() {
    m_reconnectTimer->stop();
    m_reconnectTimer->setInterval(100);

    lith()->log(Logger::Protocol, "Connected to WeeChat, starting handshake");

    QTimer::singleShot(0, lith(), &Lith::resetData);
    lith()->networkErrorStringSet(QString());

    lith()->statusSet(Lith::CONNECTED);
    QString hashAlgos;
    for (auto &i : supportedHashAlgos) {
        if (!hashAlgos.isEmpty())
            hashAlgos.append(":");
        hashAlgos.append(i);
    }

    if (lith()->settingsGet()->handshakeAuthGet()) {
        m_connection->write(QString("(%1) handshake password_hash_algo=%2,compression=%3\n").arg(MessageNames::c_handshake).arg(hashAlgos).arg(lith()->settingsGet()->connectionCompressionGet() ? "zlib" : "off").toUtf8());
    }
    else {
        StringMap data;
        data["password_hash_algo"] = "plain";
        onHandshakeAccepted(data);
    }

    m_hotlistTimer->start();
}

void Weechat::onDisconnected() {
    lith()->statusSet(Lith::DISCONNECTED);

    lith()->log(Logger::Protocol, QString("WeeChat connection lost, will reconnect in %1ms").arg(m_reconnectTimer->interval() * 2));

    m_fetchBuffer.clear();
    m_bytesRemaining = 0;
    m_hotlistTimer->stop();

    m_reconnectTimer->setInterval(m_reconnectTimer->interval() * 2);
    m_reconnectTimer->start();
}

void Weechat::onDataReceived(const QByteArray &data) {
    auto dataCopy(data);
    onMessageReceived(dataCopy);
}

void Weechat::onError(const QString &message) {
    lith()->statusSet(Lith::ERROR);
    lith()->networkErrorStringSet("Connection failed: "+ message);
}

bool Weechat::input(pointer_t ptr, const QString &data) {
    // server doesn't reply to input commands directly so no message order here
    auto line = QString("input 0x%1 %2\n").arg(ptr, 0, 16).arg(data);
    //qCritical() << "WRITING:" << line;
    auto message = line.toUtf8();
    if (message.count() == m_connection->write(line.toUtf8()))
        return true;
    return false;
}

void Weechat::fetchLines(pointer_t ptr, int count) {
    const auto* cLith = lith();
    auto *buffer = cLith->getBuffer(ptr);
    if (buffer)
        lith()->log(Logger::Protocol, buffer->nameGet(), QString("Fetching %1 lines").arg(count));
    else
        lith()->log(Logger::Unexpected, "Attempted to fetch lines for buffer with invalid buffer");
    auto line = QString("(handleFetchLines;%1) hdata buffer:0x%2/lines/last_line(-%3)/data\n").arg(m_messageOrder++).arg(ptr, 0, 16).arg(count);
    m_connection->write(line.toUtf8());
    m_timeoutTimer->start(5000);
}

void Weechat::onMessageReceived(QByteArray &data) {
    QDataStream s(&data, QIODevice::ReadOnly);

    Protocol::String id = Protocol::parse<Protocol::String>(s);

    char type[4] = { 0 };
    s.readRawData(type, 3);

    if (QString(type) == "hda") {
        Protocol::HData hda = Protocol::parse<Protocol::HData>(s);

        if (c_initializationMap.contains(id)) {
            // wtf, why can't I write this as |= ?
            m_initializationStatus = (Initialization) (m_initializationStatus | c_initializationMap.value(id, UNINITIALIZED));
            if (!QMetaObject::invokeMethod(Lith::instance(), id.toStdString().c_str(), Qt::QueuedConnection, Q_ARG(Protocol::HData, hda))) {
                lith()->log(Logger::Unexpected, QString("Possible unhandled message: %1").arg(id));
            }
        }
        else {
            auto name = id.split(";").first();
            if (!QMetaObject::invokeMethod(Lith::instance(), name.toStdString().c_str(), Qt::QueuedConnection, Q_ARG(Protocol::HData, hda))) {
                lith()->log(Logger::Unexpected, QString("Possible unhandled message: %1").arg(name));
            }
        }
    }
    else if (QString(type) == "htb") {
        Protocol::HashTable htb = Protocol::parse<Protocol::HashTable>(s);

        onHandshakeAccepted(htb);
    }
    else if (QString(type) == "str") {
        Protocol::String str = Protocol::parse<Protocol::String>(s);

        if (!QMetaObject::invokeMethod(Lith::instance(), id.toStdString().c_str(), Qt::QueuedConnection, Q_ARG(const FormattedString&, str))) {
            lith()->log(Logger::Unexpected, QString("Possible unhandled message: %1").arg(id));
        }
    }
    else {
        lith()->log(Logger::Unexpected, QString("Was not able to handle message type: %1").arg(type));
    }

    if (!s.atEnd()) {
        lith()->log(Logger::Unexpected, QString("CRITICAL! Data seemingly ended but buffer is not empty"));
    }
}

void Weechat::onPongReceived(qint64 id) {
    m_lastReceivedPong = id;
}

void Weechat::onTimeout() {
    m_connection->reset();
    lith()->statusSet(Lith::DISCONNECTED);
    start();
}

void Weechat::onPingTimeout() {
    static qint64 previousPing = 0;
    if (m_initializationStatus == COMPLETE) {
        if (previousPing < m_lastReceivedPong - 1) {
            lith()->log(Logger::Unexpected, QString("Server didn't respond to ping in time, index: %1").arg(previousPing));
            restart();
        }
        previousPing = m_messageOrder++;
        if (m_connection->write(QString("(%1) ping %1\n").arg(previousPing)) <= 0) {
            restart();
        }
    }
    else {
        //restart();
    }
}
